#!/usr/bin/env python

'''Driver UI'''

import tkinter as tk

from ac_websocket_client.objects import (
    GriddedButton, GriddedFrame, GriddedLabel, GriddedTreeview, TrafficLight)


class DriverUI(GriddedFrame):
    '''Driver UI'''

    def __init__(self, parent):

        super().__init__(grid_row=3, grid_col=0, height_by=3)

        self.parent = parent

        self.configure_columns(1, 1, 1, 1, 1, 1, 1)

        GriddedLabel(self, grid_row=0,
                     grid_col=0, width=8, text="Grid")

        self._button = tk.StringVar(value='Show Drivers')
        GriddedButton(self, grid_row=0, grid_col=2,
                      textvar=self._button, width=10,
                      command=lambda: self.parent.loop.create_task(self._toggle_grid()))
        GriddedButton(self, grid_row=0, grid_col=3,
                      text='Order 1..n', width=10,
                      command=lambda: self.parent.loop.create_task(self.parent.update_grid(by_finishing=True)))
        GriddedButton(self, grid_row=0, grid_col=4,
                      text='Order n..1', width=10,
                      command=lambda: self.parent.loop.create_task(self.parent.update_grid(by_reverse=True)))
        GriddedButton(self, grid_row=0, grid_col=5,
                      text='Show', width=10,
                      command=lambda: self.parent.loop.create_task(self.parent.update_grid()))
        GriddedButton(self, grid_row=0, grid_col=6,
                      text='Save', width=10,
                      command=lambda: self.parent.loop.create_task(self.parent.update_grid(write=True)))

        self._light = TrafficLight(self, row=0, column=7)
        self._light.gray()

        self.driver_tree = GriddedTreeview(self, 1, 0, grid_span=8)
        self.driver_tree.add_columns('Name', 'GUID', 'Car', 'Ballast',
                                     'Restrictor', 'Position', 'Connected')
        self.driver_tree.set_widths(190, 80, 190, 80, 80, 80, 80)

        self.update_ui()

    def _show_drivers(self):
        # pylint: disable=consider-using-dict-items

        self._button.set('Show Entries')
        self.driver_tree.delete(*self.driver_tree.get_children())
        for key in self.parent.drivers:
            self.driver_tree.insert('', tk.END,
                                    values=(self.parent.drivers[key]['name'],
                                            self.parent.drivers[key]['guid'],
                                            self.parent.drivers[key]['car'],
                                            self.parent.drivers[key]['ballast'],
                                            self.parent.drivers[key]['restrictor'],
                                            'n/a', 'Yes'))

    def _show_entries(self):
        # pylint: disable=consider-using-dict-items

        self._button.set('Show Drivers')
        self.driver_tree.delete(*self.driver_tree.get_children())
        for key in self.parent.entries:
            self.driver_tree.insert('', tk.END,
                                    values=(self.parent.entries[key]['drivername'],
                                            self.parent.entries[key]['guid'],
                                            self.parent.entries[key]['model'],
                                            self.parent.entries[key]['ballast'],
                                            self.parent.entries[key]['restrictor'],
                                            str(key + 1),
                                            self.parent.entries[key]['connected']))

    async def _toggle_grid(self):

        if 'Show Drivers' in self._button.get():
            self._button.set('Show Active')
        else:
            self._button.set('Show Drivers')

        self.update_ui()

    def update_ui(self):
        '''Update the UI based on states'''

        if 'Show Drivers' in self._button.get():
            self._show_drivers()
        else:
            self._show_entries()



'''Assetto Corsa Abstract Server Class'''

import aiofiles
import asyncio
from concurrent.futures import process
import hashlib
import json
import logging
import os
import pathlib
import psutil
import re
import signal
import shutil
import sys
from dataclasses import dataclass, field
from datetime import datetime
import subprocess
from typing import List


from ac_websocket_server.debug import DebugTransaction
from ac_websocket_server.error import WebsocketsServerError
from ac_websocket_server.observer import ObserverNotifier
from ac_websocket_server.protocol import Protocol

CHECK_CHILDREN_PROCESSES_ONLY = True
CHECK_PLATFORM_SPECIFIC_KARGS = False


@dataclass
class ChildServer(ObserverNotifier):
    '''Represents an Assetto Corsa childserver.'''
    # pylint: disable=logging-fstring-interpolation, invalid-name

    directory: str
    child_ini_file: str  # filename of the child's ini file
    child_title: str  # full title of child server
    child_short: str  # short descriptive name of child server
    is_optional: bool = field(init=False, default=False)

    def __post_init__(self):

        super().__init__()

        self._logger = logging.getLogger(f'ac-ws.{self.child_short}')

        self._cwd: str = self.directory
        self._exe: str   # absolute path to child executable
        self._args: tuple[str]  # arguments to use for exeuction
        self._hash: str  # md5 hash of child executable

        self._logfile_use_timestamp: bool = True
        self._logfile_stdout: str
        self._logfile_stderr: str

        self._debug_transaction = DebugTransaction(self.directory)

        self._process: asyncio.subprocess.Process

        self.orphans = False
        self.running = False    # semaphore if running
        self.stopping = False   # semaphore if stopping
        self.timestamp = ''

    def check_processes(self) -> str:
        '''Check processes and return string'''

        s = f'ac_websocket_server running with PID {os.getpid()}' + '\n'

        if self.running:
            if CHECK_CHILDREN_PROCESSES_ONLY:
                try:
                    proc = psutil.Process(pid=self._process.pid)
                    s += f'{proc.name()} running with PID {proc.pid} and parent {proc.ppid()} from {proc.cwd()}' + '\n'
                    children = psutil.Process(
                        pid=self._process.pid).children(recursive=True)
                    for proc in children:
                        s += f'{proc.name()} running with PID {proc.pid} and parent {proc.ppid()} from {proc.cwd()}' + '\n'
                except psutil.NoSuchProcess as e:
                    pass
            else:
                process_name = pathlib.PurePosixPath(self._exe).name
                for proc in psutil.process_iter(['pid', 'name', 'exe', 'cmdline', 'cwd', 'open_files']):
                    if process_name in proc.name():
                        s += f'{proc.name()} running with PID {proc.pid} and parent {proc.ppid()} from {proc.cwd()}' + '\n'
        return s

    async def check_orphans(self) -> bool:
        '''Check if orphan processes exist (from a non-controlled session)'''

        for proc in psutil.process_iter(['pid', 'name', 'exe', 'cmdline', 'cwd', 'open_files']):
            if pathlib.PurePosixPath(self._exe).name in proc.name():
                msg = f'Unable to manage AC server: {proc.name()} running with PID {proc.pid} and parent {proc.ppid()} from {proc.cwd()}' + '\n'
                self._logger.error(msg)
                await self.put(Protocol.error({'msg': msg}))
                return True

        return False

    async def log_and_send_error(self, msg: str):
        '''Send error message'''
        self._logger.error(msg)
        await self.put(Protocol.error(msg=msg))

    async def post_start_hook(self):
        '''Run after starting'''
        # pylint: disable=logging-not-lazy

        self._logger.info(f'Checking running {self._exe} processes:')
        self._logger.info('\n' + self.check_processes())

        await asyncio.sleep(5)

        for file in [self._logfile_stdout, self._logfile_stderr]:
            self._logger.info(f'Checking {file} output:')
            async with aiofiles.open(file, mode='r') as f:
                content = await f.read()
            if content:
                if re.findall(r'ERROR|RuntimeError', content):
                    self._logger.error(f'{file}:\n{content}')
                    self.running = False
                    msg = f'{self._exe} failed'
                    self._logger.error(msg)
                    await self.put(Protocol.error({'msg': msg, 'details': content}))
                else:
                    self._logger.info(f'{file}:\n{content}')

    async def post_stop_hook(self):
        '''Run after stopping'''

    def pre_start_hook(self):
        '''Run before starting'''

    def pre_stop_hook(self):
        '''Run before stopping'''
        # pylint: disable=logging-not-lazy

        self._logger.info(f'Checking running {self._exe} processes:')
        self._logger.info('\n' + self.check_processes())

    async def restart(self):
        ''''Re-start the child server'''
        self._logger.info(f'Re-starting {self.child_short}')
        await self.stop()
        await self.start()

    async def start(self, *_):
        '''Start the child server.'''
        # pylint: disable=line-too-long

        # self.orphans = await self.check_orphans()

        # if self.orphans:
        #     await self.log_and_send_error('start command failed - orphan procecess exist')
        #     return

        try:
            self.pre_start_hook()
        except WebsocketsServerError as e:
            msg = f'start command failed - {e}'
            await self.log_and_send_error(msg)
            return

        if self.running:
            msg = f'start command ignored - {self.child_short} already running'
            self._logger.info(msg)
            await self.put(Protocol.error(msg=msg))
            return

        if self._hash:
            try:
                with open(self._exe, 'rb') as file_to_check:
                    data = file_to_check.read()
                    if self._hash != hashlib.md5(data).hexdigest():
                        await self.log_and_send_error(f'{self._exe} checksum mismatch')
                        return
            except FileNotFoundError as e:
                msg = f'{self._exe} missing'
                await self.put(Protocol.error(msg=msg))
                if self.is_optional:
                    return
                else:
                    await self.log_and_send_error(f'{msg}: {e}')
                    return

        self._logger.info(f'Starting {self.child_title} server')

        os.makedirs(f'{self.directory}/logs/session', exist_ok=True)
        os.makedirs(f'{self.directory}/logs/error', exist_ok=True)

        if self._logfile_use_timestamp:
            timestamp = '-' + datetime.now().strftime("%Y%m%d_%H%M%S")
        else:
            timestamp = ''

        self._logfile_stdout = f'{self.directory}/logs/session/output{timestamp}-{self.child_title}.log'
        self._logfile_stderr = f'{self.directory}/logs/error/error{timestamp}-{self.child_title}.log'

        session_file = open(self._logfile_stdout, 'w', encoding='utf-8')
        error_file = open(self._logfile_stderr, 'w', encoding='utf-8')

        # Set platform specific options to start process in seperate process group
        kwargs = {}
        if CHECK_PLATFORM_SPECIFIC_KARGS:
            if sys.platform.startswith('win'):
                kwargs['creationflags'] = subprocess.CREATE_NEW_PROCESS_GROUP | subprocess.DETACHED_PROCESS | subprocess.CREATE_BREAKAWAY_FROM_JOB
            else:
                kwargs['start_new_session'] = True

        try:
            cmd = (self._exe,) + self._args
            self._process = await asyncio.create_subprocess_exec(
                *cmd, cwd=self._cwd, stdout=session_file, stderr=error_file, **kwargs)

            self.running = True
            self.timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")

            self._logger.info(f'Process pid is: {self._process.pid}')
            await self.put(Protocol.success(data={'msg': f'{self.child_title} started',
                                                  f'{self.child_short}': self.status_object()}))
        except FileNotFoundError as e:
            await self.log_and_send_error(f'start command failed : {e}')
            return
        except PermissionError as e:
            await self.log_and_send_error(f'start command failed : {e}')
            return

        await self.post_start_hook()

    def status_object(self):
        '''Return an object representing the child server state'''
        return {'cmd': (self._exe,) + self._args, 'cwd': self._cwd, 'timestamp': self.timestamp, 'running': self.running}

    async def status(self, *_):
        '''Check the status of the server'''
        self._logger.info('\n' + self.check_processes())
        if self.running:
            await self.put(Protocol.success(data={'msg': f'{self.child_short} is running since {self.timestamp}',
                                                  f'{self.child_short}': self.status_object()}))
        else:
            await self.put(Protocol.success(msg=f'{self.child_short} is NOT running'))

    async def stop(self, *_):
        '''Stop the child server (and any children)'''

        self.pre_stop_hook()

        if not self.running:
            msg = f'stop command ignored - {self.child_short} not running'
            self._logger.info(msg)
            await self.put(Protocol.error(msg=msg))
            return

        self._logger.info(f'Stopping {self.child_title} server')

        self.stopping = True

        try:
            children = psutil.Process(
                pid=self._process.pid).children(recursive=True)
            for child in children:
                os.kill(child.pid, signal.SIGTERM)
                self._logger.info(f'killed subprocess with PID {child.pid}')

            self._process.kill()

            status_code = await asyncio.wait_for(self._process.wait(), None)

            self._logger.info(
                f'{self.child_title} server with PID {self._process.pid} exited with status code {status_code}')
            await self.put(Protocol.success(msg=f'{self.child_title} server stopped'))
        except psutil.NoSuchProcess as e:
            self._logger.error(
                f'{self.child_title} server already exited with {e}')
            await self.put(Protocol.error(msg=f'{self.child_title} server ALREADY stopped'))
        except ProcessLookupError as e:
            self._logger.error(
                f'{self.child_title} server already exited with {e}')
            await self.put(Protocol.error(msg=f'{self.child_title} server ALREADY stopped'))

        self.running = False
        self.stopping = False

        await self.post_stop_hook()



'''Assetto Corsa Game Server Class'''

import asyncio
import configparser
import json
import logging
import os
import shutil
import sys
from dataclasses import dataclass, field
from datetime import datetime
from typing import Dict, List

from ac_websocket_server.child import ChildServer
from ac_websocket_server.entries import EntryList
from ac_websocket_server.error import WebsocketsServerError
from ac_websocket_server.grid import Grid
from ac_websocket_server.objects import (DriverInfo, EntryInfo, SessionEvent)
from ac_websocket_server.lobby import Lobby
from ac_websocket_server.protocol import Protocol
from ac_websocket_server.watcher import Watcher


@dataclass
class GameServer(ChildServer):
    '''Represents an Assetto Corsa Server.'''
    # pylint: disable=logging-fstring-interpolation, invalid-name

    directory: str
    child_ini_file: str = field(default='cfg/server_cfg.ini')
    child_title: str = field(default='AC Server')
    child_short: str = field(default='AC Server')
    is_optional: bool = False

    version: str = field(default='n/a', init=False)
    timestamp: str = field(default='n/a', init=False)
    name: str = field(init=False)
    track: str = field(init=False)
    cars: str = field(init=False)
    http_port: int = field(init=False)
    tcp_port: int = field(init=False)
    udp_port: int = field(init=False)
    drivers: Dict[str, DriverInfo] = field(init=False)
    entries: Dict[int, EntryInfo] = field(init=False)
    sessions: Dict[str, SessionEvent] = field(init=False)
    lobby: Lobby = field(init=False)
    running: int = field(init=False, default=False)

    def __post_init__(self):

        super().__post_init__()

        self._logger = logging.getLogger('ac-ws.game')
        self._args = ()

        if os.path.exists(f'{self.directory}/acServer.py'):
            self._cwd = None
            self._exe = f'{self.directory}/acServer.py'
            self._hash = None
        else:
            self._cwd = self.directory
            if sys.platform == 'linux':
                self._exe = f'{self.directory}/acServer'
                self._hash = 'f781ddfe02e68adfa170b28d0eccbbdc'
            else:
                self._exe = f'{self.directory}/acServer.exe'
                self._hash = '357e1f1fd8451eac2d567e154f5ee537'

        self.entry_list_file_name = f'{self.directory}/cfg/entry_list.ini'
        self.entry_list_backup_name = f'{self.entry_list_file_name}.old'

        self.server_cfg_file_name = f'{self.directory}/{self.child_ini_file}'
        self.server_cfg_backup_name = f'{self.server_cfg_file_name}.old'

        self.cfg = configparser.ConfigParser()
        self.cfg.optionxform = str

        self._watcher_stdout: Watcher

        self.__setup()

    async def consumer(self, message_words: List[str], connection: id = None):
        '''Consume args destined for the server'''

        message_funcs = {'drivers': self.__drivers,
                         'entries': self.__entries,
                         'info': self.__info,
                         'sessions': self.__sessions,
                         'set': self.__set,
                         'start': self.start,
                         'stop': self.stop,
                         'restart': self.restart}

        if message_funcs.get(message_words[0]):
            await message_funcs[message_words[0]](message_words[1:])

    async def __drivers(self, *_):
        '''Show game drivers info as part of a JSON reply'''
        await self.put(Protocol.success({'drivers': self.drivers}))

    async def __entries(self, *_):
        '''Show game entries info as part of a JSON reply'''
        await self.put(Protocol.success({'entries': self.entries}))

    async def __info(self, *_):
        '''Show game server info as a JSON string'''
        await self.put(Protocol.success({'server': self}))

    async def notify(self, notifier):
        '''Receive a notification of a new message from log watcher or lobby.'''

        message = await notifier.get(self)

        try:
            item = json.loads(message)

            if not item.get('type', None):
                if item.get('data', None):
                    await self.put(Protocol.success(item['data']))
                if item.get('error', None):
                    await self.put(Protocol.error(item['error']))
                return

            if item['type'] == 'LobbyEvent':
                await self.lobby.update(item['body'])
                item['body']['timestamp'] = self.lobby.since
                await self.put(Protocol.success({'lobby': {'event': item['body']}}))

            if item['type'] == 'ServerEvent':
                self.version = item['body']['version']
                self.timestamp = item['body']['timestamp']
                await self.put(Protocol.success({'server': {'event': item['body']}}))

            if item['type'] == 'DriverInfo' and item['body']['msg'] == 'joining':

                body = item['body']
                name = body['name']

                driver_info = DriverInfo()

                driver_info.name = name
                driver_info.host = body['host']
                driver_info.port = body['port']
                driver_info.car = body['car']
                driver_info.guid = body['guid']
                driver_info.ballast = body['ballast']
                driver_info.restrictor = body['restrictor']
                driver_info.msg = body['msg']

                self.drivers[driver_info.name] = driver_info

                self._logger.debug(f'Driver {name} joining')
                await self.put(Protocol.success({'driver': item['body']}))

            if item['type'] == 'DriverInfo' and item['body']['msg'] == 'leaving':
                body = item['body']
                name = body['name']
                del self.drivers[name]

                self._logger.debug(f'Driver {name} leaving')
                await self.put(Protocol.success({'driver': item['body']}))

            if item['type'] == 'SessionEvent':

                body = item['body']
                session_type = body['type']

                session_info = SessionEvent()

                for session in self.sessions:
                    self.sessions[session].active = False

                session_info.type = session_type
                session_info.laps = body['laps']
                session_info.time = body['time']
                session_info.active = True

                self.sessions[session_type] = session_info

                await self.put(Protocol.success({'session': {'event': session_info}}))

        except json.JSONDecodeError:
            pass

        # await self.put(message)

    def parse_entry_list(self):
        '''Parse entry list file and update attributes'''

        if not os.path.exists(self.entry_list_file_name):
            error_message = f'Missing entry_list.ini file in {self.directory}'
            self._logger.error(error_message)
            raise WebsocketsServerError(error_message)

        self.entry_list = EntryList(self.entry_list_file_name)
        self.entries = self.entry_list.entries

    def parse_server_cfg(self):
        '''Parse server config file and update attributes'''

        if not os.path.exists(self.server_cfg_file_name):
            error_message = f'Missing server_cfg.ini file in {self.directory}'
            self._logger.error(error_message)
            raise WebsocketsServerError(error_message)

        try:
            self.cfg.read(self.server_cfg_file_name)

            self.name = self.cfg['SERVER']['NAME']
            self.cars = self.cfg['SERVER']['CARS']
            self.track = self.cfg['SERVER']['TRACK']
            self.http_port = self.cfg['SERVER']['HTTP_PORT']
            self.tcp_port = self.cfg['SERVER']['TCP_PORT']
            self.udp_port = self.cfg['SERVER']['UDP_PORT']

            for session in ['PRACTICE', 'QUALIFY', 'RACE']:
                if self.cfg.has_section(session):
                    name = self.cfg[session].get('NAME')
                    time = self.cfg[session].get('TIME', 0)
                    laps = self.cfg[session].get('LAPS', 0)
                    self.sessions[name] = SessionEvent(
                        name, laps=laps, time=time)

        except configparser.Error as e:
            raise WebsocketsServerError(e) from e

    def pre_start_hook(self):
        '''Parse and update status before start (for re-start post __init__)'''

        super().pre_start_hook()

        self._debug_transaction.open('game-start')
        self._debug_transaction.save_file(
            f'{self.directory}/{self.child_ini_file}')
        self._debug_transaction.save_file(
            f'{self.directory}/cfg/entry_list.ini')

        timestamp = '-' + datetime.now().strftime("%Y%m%d_%H%M%S")

        shutil.copy(f'{self.directory}/cfg/entry_list.ini',
                    f'{self.directory}/logs/acws/entry_list{timestamp}-{self.child_title}.ini')
        shutil.copy(f'{self.directory}/cfg/server_cfg.ini',
                    f'{self.directory}/logs/acws/server_cfg{timestamp}-{self.child_title}.ini')

        self.grid.restore()

        try:
            self.parse_server_cfg()
            self.parse_entry_list()
            self.grid.track = self.track
        except WebsocketsServerError as e:
            self._logger.error('Command failed - configuration error')
            raise e

    async def post_start_hook(self):

        super().post_start_hook()

        await self.put(Protocol.success(
            msg='Assetto Corsa server started'))

        self.lobby.subscribe(self)

        self._watcher_stdout = Watcher(self._logfile_stdout)
        self._watcher_stdout.subscribe(self)
        await self._watcher_stdout.start()

        self._debug_transaction.close()

    async def post_stop_hook(self):

        self.__setup()

        self._debug_transaction.open('game-stop')
        self._debug_transaction.save_file(
            f'{self.directory}/{self.child_ini_file}')
        self._debug_transaction.save_file(
            f'{self.directory}/cfg/entry_list.ini')

        await self.put(Protocol.success(
            msg='Assetto Corsa server stopped'))

        await self._watcher_stdout.stop()
        self._watcher_stdout.unsubscribe(self)

        self.lobby.unsubscribe(self)

    async def __sessions(self, *_):
        '''Show game sessions info as a JSON string'''
        await self.put(Protocol.success({'sessions': self.sessions}))

    async def __set(self, args):
        '''Set server config options

        set race enabled
        set race disabled
        set race laps XX
        set race time YY

        '''

        try:
            session_name = str(args[0]).upper()
            if session_name not in ['PRACTICE', 'QUALIFY', 'RACE']:
                await self.put(Protocol.error({'msg': f'Invalid session name {session_name}'}))
                return
            option_name = str(args[1]).upper()
            if option_name not in ['ENABLE', 'DISABLE', 'LAPS', 'TIME']:
                await self.put(Protocol.error({'msg': f'Invalid option name {option_name}'}))
                return
            if option_name in ['LAPS', 'TIME']:
                option_value = args[2]
            else:
                option_value = None
        except IndexError:
            await self.put(Protocol.error({'msg': f'Failed to parse arguments to set command {args}'}))
            return

        if not os.path.exists(self.server_cfg_backup_name):
            shutil.copy(self.server_cfg_file_name, self.server_cfg_backup_name)
            await self.put(Protocol.success({'msg': f'Created {self.server_cfg_backup_name} before changes'}))
        else:
            await self.put(Protocol.success({'msg': f'Existing {self.server_cfg_backup_name} found'}))

        try:
            if option_name == 'DISABLE':
                self.cfg.remove_section(session_name)
            if option_name == 'ENABLE' and not self.cfg.has_section(session_name):
                self.cfg.add_section(session_name)
                self.cfg.set(session_name, 'NAME', session_name.capitalize())
                if session_name == 'PRACTICE' or session_name == 'QUALIFY':
                    self.cfg.set(session_name, 'TIME', '10')
                self.cfg.set(session_name, 'IS_OPEN', '1')
                if session_name == 'RACE':
                    self.cfg.set(session_name, 'LAPS', '10')
                    self.cfg.set(session_name, 'WAIT_TIME', '60')
            if option_value and option_name == 'LAPS' and session_name == 'RACE':
                self.cfg.set(session_name, option_name, option_value)
                self.cfg.remove_option(session_name, 'TIME')
            if option_value and option_name == 'TIME':
                self.cfg.set(session_name, option_name, option_value)
                self.cfg.remove_option(session_name, 'LAPS')

            with open(self.server_cfg_file_name, 'w', encoding='utf-8') as configfile:
                self.cfg.write(configfile, space_around_delimiters=False)
            await self.put(Protocol.success({'msg': f'Updated {self.server_cfg_file_name} with {args}'}))
        except IOError:
            await self.put(Protocol.error({'msg': f'Failed to update {self.server_cfg_file_name}'}))

    def __setup(self):
        '''(Re)set state'''

        self.drivers = {}
        self.entries = {}
        self.sessions = {}

        self.parse_server_cfg()
        self.parse_entry_list()

        self.grid = Grid(server_directory=self.directory,
                         track=self.track,
                         entry_list=self.entry_list)

        self.lobby = Lobby()



'''
Assetto Corsa Log Watcher.

Ideas and code from: https://github.com/kuralabs/logserver/blob/master/server/server.py
'''

import asyncio
import json
import logging
import re

import aiofiles

from ac_websocket_server.objects import DriverInfo, EnhancedJSONEncoder, LobbyEvent, ServerEvent, SessionEvent, Message, MessageType
from ac_websocket_server.observer import Notifier
from ac_websocket_server.task_logger import create_task

TAIL_DELAY = 1


class Watcher(Notifier):
    '''Represents a watcher for AC logfiles.
    Parses log files and sends messages to send_queue.'''

    def __init__(self, filename: str) -> None:
        '''Create Watcher instance for filename.'''

        self._logger = logging.getLogger('ac-ws.watcher')

        self._filename = filename

        self._readlines_task: asyncio.Task

        self._lobby_event = LobbyEvent()
        self._driver_event = DriverInfo()
        self._server_event = ServerEvent()
        self._session_event = SessionEvent()

        super().__init__()

    async def parse_lines(self, _file, lines):
        '''Parse lines of logfile and send messages to observer.'''
        # pylint: disable=invalid-name, pointless-string-statement

        for line in lines:

            '''Parse for server info - see tests/watcher_test.py for details.'''

            m = re.compile(
                r'^Assetto Corsa Dedicated Server (.*)').match(line)
            if m:
                self._server_event.version = m.group(1)
                self._server_event.msg = 'Assetto Corsa Dedicated Server starting'

            m = re.compile(
                r'^(\d{4}-\d{2}-\d{2} .*)$').match(line)
            if m:
                self._server_event.timestamp = m.group(1)

            m = re.compile(
                r'^CARS=(.*)').match(line)
            if m:
                await self.put(json.dumps(Message(type=MessageType.SERVER_EVENT,
                                                  body=self._server_event),
                                          cls=EnhancedJSONEncoder))
                continue

            '''Parse for session info - see tests/watcher_test.py for details.'''

            m = re.compile(
                r'^SENDING session name : (.*)').match(line)
            if m:
                self._session_event.type = m.group(1)
                self._session_event.msg = f'{self._session_event.type} session starting'

            m = re.compile(
                r'^SENDING session time : (.*)').match(line)
            if m:
                self._session_event.time = int(m.group(1))

            m = re.compile(
                r'^SENDING session laps : (.*)').match(line)
            if m:
                self._session_event.laps = int(m.group(1))
                await self.put(json.dumps(Message(type=MessageType.SESSION_EVENT,
                                                  body=self._session_event),
                                          cls=EnhancedJSONEncoder))
                continue

            '''Parse for driver info - see tests/watcher_test.py for details.'''

            m = re.compile(
                r'^NEW PICKUP CONNECTION from  (.*):(\d*)').match(line)
            if m:
                self._driver_event = DriverInfo()
                self._driver_event.host = m.group(1)
                self._driver_event.port = int(m.group(2))

            m = re.compile(
                r'^Looking for available slot by name for GUID (\d*) (.*)').match(line)
            if m:
                self._driver_event.guid = m.group(1)
                self._driver_event.car = m.group(2)

            m = re.compile(
                r'^Slot found at index (\d*)').match(line)
            if m:
                self._driver_event.slot = int(m.group(1))

            m = re.compile(
                r'^DRIVER: (.*) \[.*$').match(line)
            if m:
                self._driver_event.name = m.group(1)
                self._driver_event.msg = 'joining'
                await self.put(json.dumps(Message(type=MessageType.DRIVER_INFO,
                                                  body=self._driver_event),
                                          cls=EnhancedJSONEncoder))
                continue

            '''Parse for driver info - see tests/watcher_test.py for details.'''

            m = re.compile(
                r'^Clean exit, driver disconnected:  (.*) \[.*$').match(line)
            if m:
                self._driver_event = DriverInfo(name=m.group(1), msg='leaving')
                await self.put(json.dumps(Message(type=MessageType.DRIVER_INFO,
                                                  body=self._driver_event),
                                          cls=EnhancedJSONEncoder))
                continue

            '''
            CALLING http://93.57.10.21/lobby.ashx/register?name=SNRL+AC+%231&port=9601&tcp_port=9601&max_clients=12&track=rt_autodrom_most&cars=ks_mazda_mx5_cup&timeofday=-16&sessions=1,2,3&durations=7200,600,8&password=1&version=202&pickup=1&autoclutch=1&abs=1&tc=1&stability=0&legal_tyres=&fixed_setup=0&timed=0&extra=0&pit=0&inverted=0
            '''
            m = re.compile(
                r'^CALLING (.*)$').match(line)
            if m:

                self._lobby_event = LobbyEvent(url_register=m.group(1))
                await self.put(json.dumps(Message(type=MessageType.LOBBY_EVENT,
                                                  body=self._lobby_event),
                                          cls=EnhancedJSONEncoder))
                continue

            '''
            CONNECTED TO LOBBY
            '''
            m = re.compile(
                r'^CONNECTED TO LOBBY$').match(line)
            if m:
                self._lobby_event.connected = True
                if self._server_event.timestamp:
                    self._lobby_event.timestamp = self._server_event.timestamp
                await self.put(json.dumps(Message(type=MessageType.LOBBY_EVENT,
                                                  body=self._lobby_event),
                                          cls=EnhancedJSONEncoder))

            '''
            SENDING http://93.57.10.21/lobby.ashx/ping?session=2&timeleft=600&port=9601&clients=0&track=rt_autodrom_most&pickup=1
            '''
            m = re.compile(
                r'^SENDING (.*lobby.ashx\/ping.*)$').match(line)
            if m:
                self._lobby_event.url_ping = m.group(1)
                await self.put(json.dumps(Message(type=MessageType.LOBBY_EVENT,
                                                  body=self._lobby_event),
                                          cls=EnhancedJSONEncoder))
                continue
            '''
            ERROR,SERVER NOT REGISTERED WITH LOBBY - PLEASE RESTART
            ERROR - RESTART YOUR SERVER TO REGISTER WITH THE LOBBY
            '''

            m = re.compile(
                r'^ERROR.*REGISTER.*LOBBY$').match(line)
            if m:
                self._lobby_event.connected = False
                await self.put(json.dumps(Message(type=MessageType.LOBBY_EVENT,
                                                  body=self._lobby_event),
                                          cls=EnhancedJSONEncoder))

    async def readlines(self):
        '''Co-routine to be run as a task to read lines from logfile.'''
        # pylint: disable=invalid-name, logging-fstring-interpolation

        self._logger.debug(f'Started watching {self._filename}: ')

        async with aiofiles.open(self._filename, mode='r') as f:

            while True:
                line = await f.readline()
                if line:
                    await self.parse_lines(self._filename, [line])
                else:
                    await asyncio.sleep(TAIL_DELAY)

        self._logger.debug(f'Stopped watching {self._filename}: ')

    async def start(self):
        '''Start monitoring logfile'''

        self._readlines_task = create_task(self.readlines(),
                                           logger=self._logger,
                                           message='readlines() raised an exception')

    async def stop(self):
        '''Stop monitoring logfile'''

        self._readlines_task.cancel()



#!/usr/bin/env python

'''Assetto Corsa Websockets tkintr Client.

Opens a connection to a websocket server and enters an interactive loop to send/echo.

Reference: https://stackoverflow.com/questions/47895765/use-asyncio-and-tkinter-or-another-gui-lib-together-without-freezing-the-gui

'''

import argparse
import asyncio
import logging
import sys

from ac_websocket_client import DEBUG
from ac_websocket_client.app import App
from ac_websocket_server.constants import HOST, PORT


async def main(args):
    '''Parse command line arguments and start App'''

    parser = argparse.ArgumentParser(prog='client.py',
                                     description='A web-sockets test client.')

    parser.add_argument("--host",
                        type=str, default=HOST,
                        help="host to connect/listen")

    parser.add_argument("--port",
                        type=int, default=PORT,
                        help="port to connect/listen")

    parser.add_argument("--verbose",
                        action="store_true",
                        help="show verbose output")

    options = parser.parse_args(args)

    url = f'ws://{options.host}:{str(options.port)}'

    loop = asyncio.get_event_loop()

    if DEBUG:
        logging.basicConfig(level=logging.DEBUG)

    app = App(loop, url)
    await app.start_ui()


if __name__ == '__main__':

    try:
        asyncio.run(main(sys.argv[1:]))
    except RuntimeError:
        pass



#!/usr/bin/env python

'''Connection UI'''

import tkinter as tk

from ac_websocket_client.objects import (
    GriddedButton, GriddedEntry, GriddedFrame, GriddedLabel, TrafficLight)


class ConnectionUI(GriddedFrame):
    '''Connection UI'''

    def __init__(self, parent):

        super().__init__(grid_row=0, grid_col=0, height_by=0.5)

        self.parent = parent

        self.configure_columns(1, 1, 4, 1, 1)

        GriddedLabel(self, grid_row=0, grid_col=0, width=8, text="ACWS")
        GriddedLabel(self, grid_row=0, grid_col=1, width=8, text="url:")

        self.input = tk.StringVar()
        if self.parent.url:
            self.input.set(self.parent.url)

        self.entry = GriddedEntry(self, grid_row=0, grid_col=2,
                                  textvariable=self.input
                                  )
        self.entry.bind('<Return>', self._async_toggle_connection)
        self.entry.bind('<KeyRelease>', self._update_url)

        self.button = tk.StringVar(value='Connect')
        GriddedButton(self, grid_row=0, grid_col=3,
                      textvariable=self.button,
                      command=lambda: parent.loop.create_task(self.parent.toggle_connection()))

        self.light = TrafficLight(self, row=0, column=4)
        self.light.gray()

        self.update_ui()

    def _async_toggle_connection(self, _event):
        '''Connect to the websocket server'''
        self.parent.loop.create_task(self.parent.toggle_connection())

    def update_ui(self):
        '''Update the UI based on connection status'''

        self.input.set(self.parent.url)

        if self.parent.states.is_connected:
            self.button.set('Disconnect')
            self.light.green()
        else:
            self.button.set('Connect')
            self.light.red()

    def _update_url(self, _event):
        self.parent.url = self.input.get()



#!/usr/bin/env python

'''Assetto Corsa Websockets App'''

import asyncio
from concurrent.futures import TimeoutError as ConnectionTimeoutError
import json
import re
import sys
import tkinter as tk
from datetime import datetime

import websockets

from ac_websocket_client import DEBUG
from ac_websocket_client.connection import ConnectionUI
from ac_websocket_client.console import ConsoleUI
from ac_websocket_client.driver import DriverUI
from ac_websocket_client.objects import APP_TITLE
from ac_websocket_client.server import ServerUI
from ac_websocket_client.tracker import TrackerUI
from ac_websocket_server.protocol import Protocol


class App(tk.Tk):
    '''Wrapper class for Tk app'''

    UPDATE_INTERVAL = 1/120

    PROTOCOL = '0.7dev1'

    class States():
        '''Internal States of application'''
        is_connected: bool = False
        is_registered: bool = False
        is_started: bool = False
        is_tracking: bool = False

    def __init__(self, loop, url=None):
        super().__init__()

        self.loop = loop
        self.protocol("WM_DELETE_WINDOW", self.stop_ui)

        self.consumer_queue = asyncio.Queue()
        self.producer_queue = asyncio.Queue()

        self.states = App.States()

        self.drivers = {}
        self.entries = {}

        self.lobby = {}
        self.server = {}
        self.sessions = {}
        self.tracker = {}

        self.url = url
        self.websocket = None

        self.tasks = []

        self._create_ui()

    def _create_ui(self):
        '''Build the UI elements'''

        self.title(APP_TITLE)
        self.config(bg='lightgray')

        self.columnconfigure(0, weight=1)

        self.rowconfigure(0, weight=1)
        self.connection_ui = ConnectionUI(self)

        self.rowconfigure(1, weight=6)
        self.server_ui = ServerUI(self)

        self.rowconfigure(2, weight=1)
        self.tracker_ui = TrackerUI(self)

        self.rowconfigure(3, weight=4)
        self.driver_ui = DriverUI(self)

        self.rowconfigure(4, weight=4)
        self.console_ui = ConsoleUI(self)

        self.tasks.append(self.loop.create_task(self._monitor()))

        if DEBUG:
            self.update()
            self._debug_ui(self, 0)

    def _debug_ui(self, w, depth=0):
        '''Print debug information on UI'''
        print('  ' * depth + f'{w.winfo_class()} w={str(w.winfo_width())}/{str(w.winfo_reqwidth())} h={str(w.winfo_height())}/{str(w.winfo_reqheight())} x/y=+{str(w.winfo_x())}+{str(w.winfo_y())}')
        for i in w.winfo_children():
            self._debug_ui(i, depth+1)

    async def _handler(self, websocket):
        '''Handle websocket tasks'''
        consumer_task = asyncio.create_task(self._handler_consumer(websocket))
        producer_task = asyncio.create_task(self._handler_producer(websocket))
        _done, pending = await asyncio.wait(
            [consumer_task, producer_task],
            return_when=asyncio.FIRST_COMPLETED,
        )
        for task in pending:
            task.cancel()

        await self.consumer_queue.put(Protocol.success(data={'sent': {'websocket.cancel()': self.url}}))

    async def _handler_consumer(self, websocket):
        '''Handle messages received from websocket'''
        async for msg in websocket:
            await self.consumer_queue.put(msg)

    async def _handler_producer(self, websocket):
        '''Handle messages received to send on websocket'''
        while True:
            try:
                message = await self.producer_queue.get()
                await websocket.send(message)
                await self.consumer_queue.put(Protocol.success(data={'sent': message}))
            except Exception:
                print('\n> Connection Closing', file=sys.stderr)
                return

    async def _monitor(self):
        '''Monitor incoming messages and send to connection listbox'''
        while True:
            try:
                input_line = json.loads(await self.consumer_queue.get())

                output_timestamp = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
                output_lines = ''

                if error_msg := input_line.get('error', None):

                    if error_msg['msg'] == 'ERROR,INVALID SERVER,CHECK YOUR PORT FORWARDING SETTINGS':
                        self.states.is_registered = False
                        self.server_ui.update_ui()

                    output_lines = json.dumps(error_msg, indent=4)
                    self.console_ui.output(
                        output_timestamp, output_lines, fg='Red')
                    continue

                if success_msg := input_line.get('data', None):

                    if success_msg.get('sent', None):

                        output_lines = json.dumps(success_msg, indent=4)
                        self.console_ui.output(
                            output_timestamp, output_lines, fg='Black')
                        continue

                    if driver_msg := success_msg.get('driver', None):
                        if guid := driver_msg.get('guid', None):
                            if driver_joining := driver_msg.get('msg', None):
                                slot = driver_msg['slot']
                                self.entries[slot]['guid'] = guid
                                self.entries[slot]['drivername'] = driver_msg['name']
                                if 'joining' in driver_joining:
                                    self.drivers[guid] = driver_msg
                                    self.entries[slot]['connected'] = 'Yes'
                                else:
                                    self.entries[slot]['connected'] = 'No'
                                    del self.drivers[guid]
                                self.driver_ui.update_ui()

                    if lobby := success_msg.get('lobby', None):
                        self.states.is_registered = self.lobby['event']['connected']
                        self.lobby.since = self.lobby['event']['timestamp']
                        self.server_ui.update_ui()

                    if server_msg := success_msg.get('server', None):
                        if server_event := server_msg.get('event'):
                            self.server['timestamp'] = server_event['timestamp']
                            self.server['version'] = server_event['version']
                            self.states.is_started = True
                        else:
                            self.server = server_msg
                            self.states.is_started = self.server.get(
                                'running', False)
                        self.server_ui.update_ui()
                        if entries := server_msg.get('entries'):
                            self.entries = {}
                            for entry in entries:
                                slot = int(re.sub(r'\D', '', entry))
                                self.entries[slot] = entries[entry]
                                self.entries[slot]['connected'] = 'No'
                        if lobby := server_msg.get('lobby', None):
                            self.lobby = lobby
                            self.states.is_registered = self.lobby['connected']
                        self.driver_ui.update_ui()

                    if generic_msg := success_msg.get('msg', None):
                        if generic_msg == 'Stracker started' or re.search(r'^.*stracker.*is running.*$', generic_msg):
                            self.states.is_tracking = True
                            self.tracker_ui.update_ui()
                        if generic_msg == 'Stracker server stopped':
                            self.states.is_tracking = False
                            self.tracker_ui.update_ui()
                        if generic_msg == 'Assetto Corsa server started':
                            self.states.is_started = True
                            self.server_ui.update_ui()
                        if generic_msg == 'Assetto Corsa server stopped':
                            self.states.is_started = False
                            self.server_ui.update_ui()

                    if tracker_msg := success_msg.get('stracker', None):
                        self.tracker = tracker_msg
                        self.states.is_tracking = self.tracker['running']
                        self.tracker_ui.update_ui()

                    if sessions_msg := success_msg.get('session', None):
                        if session_event := sessions_msg.get('event'):
                            session_type = session_event['type']
                            self.sessions[session_type]['active'] = session_event['active']
                            self.server_ui.update_ui()

                    if sessions_msg := success_msg.get('sessions', None):
                        self.sessions = sessions_msg
                        self.server_ui.update_ui()

                    output_lines = json.dumps(success_msg, indent=4)
                    self.console_ui.output(
                        output_timestamp, output_lines, fg='Green')

                else:
                    output_lines = json.dumps(input_line, indent=4)
                    self.console_ui.output(
                        output_timestamp, output_lines, fg='Purple')

            except Exception as err:
                print(
                    f'Exception: "{err}" caught when processing:\n{input_line}')

    async def send_command(self):
        '''Send command to ACWS server'''

        if self.states.is_connected:
            if command := self.console_ui.input():
                await self.consumer_queue.put(Protocol.success(data={'sent': command}))
                await self.producer_queue.put(command)
        else:
            await self.consumer_queue.put(Protocol.error(msg='Not connected to ACWS server'))

    async def start_ui(self, interval=UPDATE_INTERVAL):
        '''Start a the update of the UI'''
        while True:
            self.update()
            await asyncio.sleep(interval)

    def stop_ui(self):
        '''Cleanup all tasks'''
        for task in self.tasks:
            task.cancel()
        self.loop.stop()
        self.destroy()

    async def toggle_connection(self):
        '''Connect to the websocket server'''

        if self.states.is_connected:
            await self.websocket.close()
            return

        if not self.url:
            return

        try:
            await self.consumer_queue.put(Protocol.success(data={'sent': {'websocket.connect()': self.url}}))
            websocket = await asyncio.wait_for(websockets.connect(self.url), 10)
            self.title(f'{APP_TITLE} - Connected to {self.url}')
            self.states.is_connected = True
            self.connection_ui.update_ui()
            self.websocket = websocket
            await self.producer_queue.put(f'protocol {App.PROTOCOL}')
            await self.producer_queue.put('server info')
            await self.producer_queue.put('server sessions')
            await self.producer_queue.put('tracker status')
            await self._handler(websocket)
            await self.consumer_queue.put(Protocol.success(msg=f'Disconnecting from {self.url}'))
            self.websocket = None
            self.title(f'{APP_TITLE}')
            self.states.is_connected = False
            self.states.is_started = False
            self.states.is_tracking = False
            self._update_ui()
        except ConnectionTimeoutError as _error:
            await self.consumer_queue.put(Protocol.error(msg=str(f'Timeout connecting to {self.url}')))
        except OSError as error:
            await self.consumer_queue.put(Protocol.error(msg=str(error)))

    async def toggle_game(self):
        '''Start the game'''

        if self.states.is_connected:
            if not self.states.is_started:
                await self.producer_queue.put('server start')
                await self.producer_queue.put('server info')
            else:
                await self.producer_queue.put('server stop')
                await self.producer_queue.put('server info')
        else:
            await self.consumer_queue.put(Protocol.error(msg='Not connected to ACWS server'))

        self.server_ui.update_ui()

    async def toggle_registration(self):
        '''(Re)-register in lobby'''

        if self.states.is_connected:
            await self.producer_queue.put('lobby restart')
        else:
            await self.consumer_queue.put(Protocol.error(msg='Not connected to ACWS server'))

        self.server_ui.update_ui()

    async def toggle_tracker(self):
        '''Toggle tracker'''

        if self.states.is_connected:
            if not self.states.is_tracking:
                await self.producer_queue.put('tracker start')
            else:
                await self.producer_queue.put('tracker stop')
        else:
            await self.consumer_queue.put(Protocol.error(msg='Not connected to ACWS server'))

        self.tracker_ui.update_ui()

    async def update_grid(self,
                          by_finishing: bool = None,
                          by_reverse: bool = None, write: bool = None):
        '''Set the grid order'''

        if self.states.is_connected:
            if by_finishing:
                await self.producer_queue.put('grid finish')
                return
            if by_reverse:
                await self.producer_queue.put('grid reverse')
                return
            if write:
                await self.producer_queue.put('grid save')
                return
            await self.producer_queue.put('grid entries')
        else:
            await self.consumer_queue.put(Protocol.error(msg='Not connected to ACWS server'))

    def _update_ui(self):
        '''Update all UI elements'''

        self.connection_ui.update_ui()
        self.console_ui.update_ui()
        self.driver_ui.update_ui()
        self.server_ui.update_ui()
        self.tracker_ui.update_ui()



#!/usr/bin/env python

'''Tracker UI'''

import tkinter as tk

from ac_websocket_client.objects import (
    GriddedButton, GriddedEntry, GriddedFrame, GriddedLabel, TrafficLight)


class TrackerUI(GriddedFrame):
    '''Tracker UI'''

    def __init__(self, parent):

        super().__init__(grid_row=2, grid_col=0, height_by=0.5)

        self.parent = parent

        self.configure_columns(1, 1, 4, 1, 1)

        GriddedLabel(self, grid_row=0,
                     grid_col=0, width=8, text="Tracker")
        GriddedLabel(self, grid_row=0,
                     grid_col=1, width=8, text="started:")

        self._field = tk.StringVar()
        GriddedEntry(self, grid_row=0, grid_col=2,
                     textvariable=self._field, state=tk.DISABLED)

        self._button = tk.StringVar(value='Start Tracker')
        GriddedButton(self, grid_row=0, grid_col=3,
                      textvariable=self._button,
                      command=lambda: self.parent.loop.create_task(self.parent.toggle_tracker()))
        self._light = TrafficLight(self, row=0, column=4)

        self.update_ui()

    def update_ui(self):
        '''Update the UI based on states'''

        if self.parent.states.is_tracking:
            self._light.green()
            if timestamp := self.parent.tracker.get('timestamp', None):
                self._field.set(timestamp)
            else:
                self._field.set('')
            self._button.set('Stop Tracker')
        else:
            self._light.red()
            self._field.set('n/a')
            self._button.set('Start Tracker')



#!/usr/bin/env python

'''Console UI'''

import tkinter as tk

from ac_websocket_client.objects import (
    GriddedButton, GriddedEntry, GriddedFrame,
    GriddedLabel, GriddedListbox, GriddedText, TrafficLight)

USE_TEXT = False


class ConsoleUI(GriddedFrame):
    '''Console UI'''

    def __init__(self, parent):

        super().__init__(grid_row=4, grid_col=0, height_by=3)

        self.parent = parent

        self.configure_columns(1, 1, 4, 1, 1)

        GriddedLabel(self, grid_row=0, grid_col=0, width=8, text="Console")
        GriddedLabel(self, grid_row=0, grid_col=1, width=8, text="command:")

        self._field = tk.StringVar()
        self.entry = GriddedEntry(self, grid_row=0, grid_col=2,
                                  textvariable=self._field)
        self.entry.bind('<Return>', self._async_send_command)

        self._button = tk.StringVar(value='Send')
        GriddedButton(self, grid_row=0, grid_col=3,
                      textvariable=self._button,
                      command=lambda: self.parent.loop.create_task(self.parent.send_command()))

        self._light = TrafficLight(self, row=0, column=4)
        self._light.gray()

        if USE_TEXT:
            self._text = GriddedText(
                self, grid_row=1, grid_col=0, grid_span=5)
        else:
            self._listbox = GriddedListbox(
                self, grid_row=1, grid_col=0, grid_span=5)

        self.update_ui()

    def _async_send_command(self, _event):
        '''Send command to ACWS server'''
        self.parent.loop.create_task(self.parent.send_command())

    def input(self):
        '''Return command input to be sent to ACWS server'''
        return self._field.get()

    def output(self, timestamp: str, lines: str, fg='Black'):
        '''Send command output to console.'''

        output_fmt = {'fg': fg}

        for line in lines.splitlines():
            if USE_TEXT:
                self._text.insert(tk.END, f'{timestamp}: {line}\n')
            else:
                self._listbox.insert(tk.END, f'{timestamp}: {line}\n')
                self._listbox.itemconfig(tk.END, output_fmt)

        self.update_ui()

    def update_ui(self):
        '''Update the UI'''

        if USE_TEXT:
            self._text.yview(tk.END)
        else:
            self._listbox.yview(tk.END)



#!/usr/bin/env python

'''Server UI'''

import tkinter as tk

from ac_websocket_client.objects import (
    GriddedButton, GriddedEntry, GriddedFrame, GriddedLabel, TrafficLight)


class ServerUI(GriddedFrame):
    '''Server UI'''

    def __init__(self, parent):

        super().__init__(grid_row=1, grid_col=0, height_by=3)

        self.parent = parent

        self.configure_columns(1, 1, 4, 1, 1)

        self._buttons = {}
        self._fields = {}
        self._lights = {}

        GriddedLabel(self, grid_row=0, grid_col=0, width=8, text="Game")
        GriddedLabel(self, grid_row=0, grid_col=1, width=8, text="started:")

        self._fields['started'] = tk.StringVar()
        GriddedEntry(self, grid_row=0, grid_col=2,
                     textvariable=self._fields['started'], state=tk.DISABLED)

        self._fields['registered'] = tk.StringVar()
        GriddedLabel(self, grid_row=1, grid_col=1,
                     width=8, text="registered:")
        GriddedEntry(self, grid_row=1, grid_col=2,
                     textvariable=self._fields['registered'], state=tk.DISABLED)

        self._fields['name'] = tk.StringVar()
        GriddedLabel(self, grid_row=2, grid_col=1, width=8, text="name:")
        GriddedEntry(self, grid_row=2, grid_col=2,
                     textvariable=self._fields['name'], state=tk.DISABLED)

        self._fields['track'] = tk.StringVar()
        GriddedLabel(self, grid_row=3, grid_col=1, width=8, text="track:")
        GriddedEntry(self, grid_row=3, grid_col=2,
                     textvariable=self._fields['track'], state=tk.DISABLED)

        self._fields['cars'] = tk.StringVar()
        GriddedLabel(self, grid_row=4, grid_col=1, width=8, text="cars:")
        GriddedEntry(self, grid_row=4, grid_col=2,
                     textvariable=self._fields['cars'], state=tk.DISABLED)

        self._fields['sessions'] = {}
        self._lights['sessions'] = {}

        self._fields['sessions']['Practice'] = tk.StringVar()
        GriddedLabel(self, grid_row=5, grid_col=1,
                     width=8, text="practice session:")
        GriddedEntry(self, grid_row=5, grid_col=2,
                     textvariable=self._fields['sessions']['Practice'], state=tk.DISABLED)
        self._lights['sessions']['Practice'] = TrafficLight(
            self, row=5, column=4)
        self._lights['sessions']['Practice'].gray()

        self._fields['sessions']['Qualify'] = tk.StringVar()
        GriddedLabel(self, grid_row=6, grid_col=1,
                     width=8, text="qualify session:")
        GriddedEntry(self, grid_row=6, grid_col=2,
                     textvariable=self._fields['sessions']['Qualify'], state=tk.DISABLED)
        self._lights['sessions']['Qualify'] = TrafficLight(
            self, row=6, column=4)
        self._lights['sessions']['Qualify'].gray()

        self._fields['sessions']['Race'] = tk.StringVar()
        GriddedLabel(self, grid_row=7, grid_col=1,
                     width=8, text="race session:")
        GriddedEntry(self, grid_row=7, grid_col=2,
                     textvariable=self._fields['sessions']['Race'], state=tk.DISABLED)
        self._lights['sessions']['Race'] = TrafficLight(self, row=7, column=4)
        self._lights['sessions']['Race'].gray()

        self._buttons['game'] = tk.StringVar(value='Start Game')
        GriddedButton(self, grid_row=0, grid_col=3,
                      textvariable=self._buttons['game'],
                      command=lambda: self.parent.loop.create_task(self.parent.toggle_game()))
        self._lights['game'] = TrafficLight(self, row=0, column=4)

        self._buttons['lobby'] = tk.StringVar(value='(Re)register')
        GriddedButton(self, grid_row=1, grid_col=3,
                      textvariable=self._buttons['lobby'],
                      command=lambda: self.parent.loop.create_task(self.parent.toggle_registration()))
        self._lights['lobby'] = TrafficLight(self, row=1, column=4)

        self.update_ui()

    def update_ui(self):
        '''Update the UI with the contents of the parent.server'''

        self._fields['name'].set(self.parent.server.get('name', ''))
        self._fields['track'].set(self.parent.server.get('track', ''))
        self._fields['cars'].set(self.parent.server.get('cars', ''))

        if not self.parent.states.is_connected:
            self._lights['game'].gray()
            self._lights['lobby'].gray()
            for session in ('Practice', 'Qualify', 'Race'):
                if self.parent.sessions.get(session, None):
                    self.parent.sessions[session]['active'] = False
                    self._lights['sessions'][session].gray()

        if self.parent.states.is_started:
            self._fields['started'].set(
                self.parent.server.get('timestamp', None))
            self._buttons['game'].set('Stop Game')
            self._lights['game'].green()
        else:
            self._fields['started'].set('n/a')
            self._buttons['game'].set('Start Game')
            self._lights['game'].red()

        if self.parent.states.is_registered and self.parent.states.is_started:
            self._fields['registered'].set(
                self.parent.lobby.get('since', ''))
            self._buttons['lobby'].set('Re-register')
            self._lights['lobby'].green()
        else:
            self._fields['registered'].set('n/a')
            self._buttons['lobby'].set('Register')
            self._lights['lobby'].red()

        if sessions := self.parent.sessions:
            for session_type in sessions:
                if session_type not in ('Practice', 'Qualify', 'Race'):
                    return
                session_active = sessions[session_type]['active']
                if session_active:
                    self._lights['sessions'][session_type].green()
                else:
                    self._lights['sessions'][session_type].gray()
                session_description = str(
                    sessions[session_type]['time'])
                if sessions[session_type]['laps'] == 0:
                    session_description += ' minutes'
                else:
                    session_description += ' laps'
                self._fields['sessions'][session_type].set(session_description)



'''Asyncio Debug Utilities.'''

import asyncio
from datetime import datetime
import os
import shutil

DELAY = 10


class DebugMonitor:
    '''Helper class to monitor async tasks'''

    @classmethod
    async def monitor_tasks(cls):
        '''
        Monitor tasks and print out every DELAY seconds.

        Call with:
        loop.create_task(monitor_tasks())
        '''
        # pylint: disable=expression-not-assigned

        while True:
            print('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>')
            tasks = [
                t for t in asyncio.all_tasks()
                if t is not asyncio.current_task()
            ]
            [t.print_stack(limit=5) for t in tasks]
            await asyncio.sleep(DELAY)


class DebugTransaction():
    '''Super-class with methods to store debug files'''

    def __init__(self, server_directory: str):

        self._server_directory = server_directory

        if os.path.isdir(f'{self._server_directory}/debug'):
            self.debug_directory = f'{self._server_directory}/debug'
        else:
            self.debug_directory = None

        self.debug_transaction = None

    def close(self):
        '''Close a debug transaction'''
        self.debug_transaction = None

    def open(self, name: str = 'debug'):
        '''Open a debug transaction'''

        if not self.debug_transaction and self.debug_directory:
            timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
            self.debug_transaction = f'{self.debug_directory}/{timestamp}-{name}'
            os.mkdir(self.debug_transaction)

    def save_file(self, filename: str, new_filename: str = None):
        '''Save a file in the debug transaction'''

        if self.debug_transaction:
            try:
                if new_filename:
                    shutil.copy2(
                        filename, '{self.debug_transaction}/{new_filename}')
                else:
                    shutil.copy2(filename, self.debug_transaction)
            except OSError as error:
                pass



'''Assetto Corsa entry_list.ini helper Class to re-order and re-write'''

import configparser
import copy
import json
from enum import Enum, auto
import logging
import random
import sys
from typing import Any, Dict, List


from ac_websocket_server.objects import EntryInfo
from ac_websocket_server.error import WebsocketsServerError

EXTRA_DEBUG = True


class EntryListIterationMethod(Enum):
    '''Iteration method to rewrite file in various manners.'''
    ORIGINAL = auto()
    FINISHING = auto()
    REVERSE = auto()
    RANDOM = auto()


class EntryList:
    '''A collection of individual Entry for the entry_list.ini file'''

    def __init__(self, file_name: str = None, entries: Dict[int, EntryInfo] = None) -> None:
        '''
        Create a new EntryList with optional input file and pre-populated entries.
        '''

        self.__logger = logging.getLogger('ac-ws.entries')

        self.file_name = file_name

        if entries:
            self.entries = entries
        else:
            self.entries = {}

        self.entries_by_entries = list()

        if file_name and not entries:
            self.parse_entries_file()

        self.results = {}

        self.results_by_entries = list()
        self.results_by_missing = list()
        self.results_by_finish = list()

        self.results_by_order = list()

        self.iteration_order = EntryListIterationMethod.ORIGINAL

    def parse_entries_file(self):
        '''Parse the original entries file.'''

        if self.file_name:

            try:
                config = configparser.ConfigParser()
                config.read(self.file_name)

                for car_id in config.sections():

                    car = config[car_id]

                    idx = int(car_id.replace('CAR_', ''))

                    self.entries[idx] = \
                        EntryInfo(car_id=car_id,
                                  model=car.get('MODEL', ''),
                                  skin=car.get('SKIN', ''),
                                  spectator_mode=car.get('SPECTATOR_MODE', ''),
                                  drivername=car.get('DRIVERNAME', ''),
                                  team=car.get('TEAM', ''),
                                  guid=car.get('GUID', ''),
                                  ballast=car.get('BALLAST', ''),
                                  restrictor=car.get('RESTRICTOR', '')
                                  )

                    self.entries_by_entries.append(idx)

            except configparser.Error as error:
                raise WebsocketsServerError(error) from error

        if EXTRA_DEBUG:
            print(f'\nentries_by_entries = {self.entries_by_entries}')

    def parse_result_file(self, result_file: str, track: str = None):
        '''Parse AC race results file'''
        # pylint: disable=invalid-name, logging-fstring-interpolation

        try:
            with open(result_file, 'r', encoding='UTF-8') as f:
                data = json.load(f)
        except OSError as error:
            self.__logger.error(f'Unable to read input file: {result_file}')
            raise OSError from error

        if track and track != data["TrackName"]:
            raise WebsocketsServerError(
                f'TrackName from {result_file} does not match {track}')

        self.results_by_finish = list()

        cars = data["Cars"]

        for car in cars:

            car_id = car['CarId']

            self.results[car_id] = \
                EntryInfo(car_id=car_id,
                          model=car['Model'],
                          skin=car['Skin'],
                          spectator_mode=self.entries[car_id].spectator_mode,
                          drivername=car['Driver']['Name'],
                          team=car['Driver']['Team'],
                          guid=car['Driver']['Guid'],
                          ballast=car['BallastKG'],
                          restrictor=car['Restrictor']
                          )

            self.results_by_entries.append(car_id)

        results = data["Result"]

        for result in results:

            car_id = result['CarId']
            total_time = result['TotalTime']

            try:
                if total_time > 0:
                    self.results_by_finish.append(car_id)
                else:
                    self.results_by_missing.append(car_id)
            except KeyError as error:
                msg = f'Entry for {error} missing from entry.ini'
                self.__logger.error(msg=msg)
                raise WebsocketsServerError(msg) from error

        all_entries_in_results = set(
            self.results_by_finish + self.results_by_missing)
        all_entries_in_entries = set(self.entries_by_entries)
        set_of_missing_entries = all_entries_in_entries.difference(
            all_entries_in_results)

        for missing in set_of_missing_entries:
            self.results[missing] = copy.deepcopy(self.entries[missing])
            self.results_by_missing.append(missing)

        if EXTRA_DEBUG:
            print(f'results_by_finish = {self.results_by_finish}')
            print(f'results_by_missing = {self.results_by_missing}')

    def __repr__(self):
        '''Create a string representation based on IterationMethod'''

        self.__set_output_order()

        entries = ""

        position = 0
        for car_id in self.results_by_order:
            self.results[car_id].new_id = position
            entries += str(self.results[car_id])
            position += 1

        return entries

    def __set_output_order(self):
        '''Set the output order based on iteration method'''

        if self.iteration_order == EntryListIterationMethod.ORIGINAL:
            self.results_by_order = self.entries_by_entries

        if self.iteration_order == EntryListIterationMethod.FINISHING:
            self.results_by_order = self.results_by_finish + self.results_by_missing

        if self.iteration_order == EntryListIterationMethod.REVERSE:
            results_by_reversed = copy.deepcopy(self.results_by_finish)
            results_by_reversed.reverse()
            self.results_by_order = results_by_reversed + self.results_by_missing

        if self.iteration_order == EntryListIterationMethod.RANDOM:
            raise ValueError('Not Implemented')

    def set_original_order(self):
        '''Set iteration to original cars order from entry list'''
        self.iteration_order = EntryListIterationMethod.ORIGINAL

    def set_standard_order(self):
        '''Set iteration to finishing order form last race'''
        self.iteration_order = EntryListIterationMethod.FINISHING

    def set_random_order(self):
        '''Set iteration to random'''
        self.iteration_order = EntryListIterationMethod.RANDOM

    def set_reversed_order(self):
        '''Set iteration to reversed grid from last race'''
        self.iteration_order = EntryListIterationMethod.REVERSE

    def show_entries(self) -> str:
        '''Returns a dictionary representation of all cars from entry_list.ini'''

        original_iteration_method = self.iteration_order
        self.iteration_order = EntryListIterationMethod.ORIGINAL

        self.__set_output_order()

        entries = {}

        position = 1
        for car_id in self.results_by_order:
            entry = self.entries[car_id]
            entry.new_id = position
            if entry.drivername is not None:
                driver_details = f'{entry.drivername} ({entry.guid})'
            else:
                driver_details = f'{entry.drivername}'
            entries[str(position)] = f'{entry.model}{driver_details}'
            position += 1

        self.iteration_order = original_iteration_method

        return entries

    def show_grid(self) -> Dict[str, Any]:
        '''Returns a dictionary representation of the grid by player'''

        self.__set_output_order()

        entries = {}

        position = 1
        for car_id in self.results_by_order:
            if car_id not in self.results_by_missing:
                if not self.results:
                    entries[position] = self.entries[car_id].drivername
                else:
                    entries[position] = self.results[car_id].drivername
                position += 1

        return entries

    def write(self, output_file):
        '''Write updated entry list to output_file'''
        # pylint: disable=invalid-name, logging-fstring-interpolation

        try:
            with open(output_file, 'w', encoding='utf-8') as f:
                f.write(str(self))
                f.close()
        except OSError as err:
            self.__logger.error(f'Unable to write output file: {output_file}')
            raise OSError from err

        if EXTRA_DEBUG:
            print('entry_list.ini:')
            print(str(self))



'''Generic observer/notifier super-class.'''

import asyncio
import logging
from typing import Any

VERBOSE_DEBUG = False


class Observer:
    '''Super class for an Observer class.

    The observer must use notifer.subscribe() and
    notifier.unsubscribe() messages to ensure that
    notify() messages will be received.'''

    def __init__(self):
        '''Initialise an Observer with a local queue.'''

        self._notifier_queue: asyncio.Queue = asyncio.Queue()

        super().__init__()

    async def notify(self, notifier):
        '''Receive a notification of a new message from
        notifier instance.  Pull the data off the notifier's
        queue and store in the local queue.'''

        message = await notifier.get(self)
        await self._notifier_queue.put(message)


class Notifier:
    '''Super class for a Notifier class.

    The notifier receives subscribe() and unsubscribe()
    messages to register interested observers.

    The notifier sends a notify() to observers when
    data exists and then receives a get() to pop the
    data from queue.'''

    def __init__(self):
        '''Initialise a Notifier with a local queue.'''

        self._logger = logging.getLogger('ac-ws.notifier')

        self._observer = {}
        self._observer_queues = {}

        super().__init__()

    async def get(self, observer: Any):
        '''
        Fetch an item from the queue.  Returns None if the queue is empty.
        Uses the observer value to index the correct queue.
        '''
        # pylint: disable=logging-fstring-interpolation

        try:
            response = self._observer_queues[id(observer)].get_nowait()
        except asyncio.QueueEmpty:
            response = None

        if VERBOSE_DEBUG:
            self._logger.debug(f'get({observer}) -> {response})')

        return response

    async def put(self, item, observer: Any = None):
        '''
        Put an item on a specific local queue and notify observer
        (if given) or all observers if None'''
        # pylint: disable=consider-using-dict-items, consider-iterating-dictionary, logging-fstring-interpolation

        if VERBOSE_DEBUG:
            self._logger.debug(f'put(self, {item}, {observer})')

        if observer:
            await self._observer_queues[id(observer)].put(item)
            await self._observer[id(observer)].notify(self)
        else:
            for key in self._observer_queues.keys():
                await self._observer_queues[key].put(item)
            for key in self._observer.keys():
                await self._observer[key].notify(self)

    def subscribe(self, observer):
        '''Subscribe an observer object for state changes.   Create a new queue per observer.
        Observer object must include an async notify(self, observable, *args, **kwargs) method.'''
        self._observer_queues[id(observer)] = asyncio.Queue()
        self._observer[id(observer)] = observer

    def unsubscribe(self, observer):
        '''Unsubscribe an observer object.'''
        try:
            del self._observer_queues[id(observer)]
            del self._observer[id(observer)]
        except ValueError:
            pass


class ObserverNotifier(Observer, Notifier):
    '''Super class that is both an Observer and Notifier.'''

    def __init__(self):
        # pylint: disable=useless-super-delegation
        super().__init__()



'''Assetto Corsa Lobby Class'''

import aiohttp
import asyncio
import logging
import os
import sys
from dataclasses import dataclass, field
from datetime import datetime
from typing import Dict, List

from ac_websocket_server.objects import LobbyEvent
from ac_websocket_server.observer import Notifier
from ac_websocket_server.protocol import Protocol


@dataclass
class Lobby(Notifier):
    '''Represents an Assetto Corsa Lobby'''

    connected: bool = field(default=False, init=False)
    since: datetime = field(default=None, init=False)
    url_register: str = field(default=None, init=False)
    url_ping: str = field(default=None, init=False)

    def __post_init__(self):

        super().__init__()

        self._logger = logging.getLogger('ac-ws.lobby')

    async def consumer(self, message_words: List[str], connection: id = None):
        '''Consume args destined for the lobby'''

        message_funcs = {'info': self._info,
                         'restart': self._restart}

        if message_funcs.get(message_words[0]):
            await message_funcs[message_words[0]]()

    async def _info(self):
        '''Send lobby information'''
        await self.put(Protocol.success({'lobby': self}))

    async def _restart(self):
        '''Re-start the lobby connection'''

        self._logger.info('Re-regestering to lobby')
        async with aiohttp.ClientSession() as session:

            async with session.get(self.url_register) as resp:
                text = await resp.text()
                if resp.status == 200 and 'ERROR' not in text:
                    self._logger.info('Successfully re-registered to lobby')
                    await self.put(Protocol.success(msg=text))
                else:
                    self._logger.error(
                        f'Failed attempt to re-register with lobby: {text}')
                    await self.put(Protocol.error(msg=text))

    async def update(self, lobby_event: LobbyEvent):
        '''Update lobby info'''

        if lobby_event.connected and not self.connected:
            self.connected = True
            self.since = lobby_event.get(
                'timestamp', datetime.now().strftime())

        if not lobby_event.connected and self.connected:
            self.connected = False
            self.since = ''
            await self._restart()

        if lobby_event.url_register:
            self.url_register = lobby_event.url_register

        if lobby_event.url_ping:
            self.url_ping = lobby_event.url_ping
